from pathlib import Path
from typing import Union

import requests  # type:ignore
import yaml  # type:ignore
from rich import print
from rich.progress import Progress


def load_yaml(filename: Union[str, Path]):  # pragma: no cover
    with open(filename, "r") as f:
        return yaml.safe_load(f)


def write_yaml(data: dict, filename: Union[str, Path]):  # pragma: no cover
    with open(filename, "w") as f:
        yaml.dump(data, f)


def url_download(  # pragma: no cover
    url: str, filename: Union[str, Path, None] = None, block_size: int = 1024, **kwargs
) -> None:
    """Downloads a file to a specified path.

    Args:
        url: The URL to download.
        filename: The path to download the file to.
        block_size: Buffer size in bytes for sending a file-like message body.
        **kwargs: Keyword arguments are passed to 'requests'

    Raises:
        HttpError: If the request response is not 200 and OK.
    """
    try:
        response = requests.get(url, stream=True, allow_redirects=True, **kwargs)
        response.raise_for_status()

        total_content_length = int(response.headers.get("content-length", 0))
        if filename is None:
            filename = url.split("/")[-1]

        with Progress(refresh_per_second=1500) as progress:
            task = progress.add_task("[red]Downloading...", total=total_content_length)

            with open(filename, "wb") as file:
                for data in response.iter_content(block_size):
                    file.write(data)
                    progress.update(task, advance=block_size)

    except requests.exceptions.HTTPError as err:
        print(err)



from typing import Optional

import pandas as pd
from cached_property import cached_property

from .._normalize import NormalizeColumns
from .._settings import s3_bionty_assets
from .._table import EntityTable

FILENAMES = {
    "human": "5WBmdkTO4JCFzPzBcDOJ3.parquet",
    "mouse": "6vgntdGiAbz5bEYP53sma.parquet",
}


def _get_shortest_name(df, column, new_column="name"):
    """Get a single shortest name from a column of lists."""
    name_list = []
    names_list = []
    for i in df[column]:
        i = i.replace(", ", "|")
        names_list.append(i)

        def shortest_name(lst: list):
            return min(lst, key=len)

        names = i.split("|")
        no_space_names = [i for i in names if " " not in i]
        if len(no_space_names) == 0:
            name = shortest_name(names)
        else:
            name = shortest_name(no_space_names)
        name_list.append(name)

    df[new_column] = name_list
    df[column] = names_list


class Protein(EntityTable):
    """Protein.

    Args:
        species: `name` of `Species` entity EntityTable.
    """

    def __init__(
        self,
        species: str = "human",
        id: Optional[str] = None,
        database: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None:
        super().__init__(id=id, database=database, version=version)
        if FILENAMES.get(species) is None:
            raise NotImplementedError
        self._species = species
        self._id_field = "uniprotkb_id" if id is None else id

    @property
    def species(self):
        """The `name` of `Species` entity EntityTable."""
        return self._species

    @cached_property
    def df(self):
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/uniprot-protein
        """
        cloudpath = s3_bionty_assets(FILENAMES.get(self.species))
        self._filepath = cloudpath.fspath

        df = pd.read_parquet(self._filepath)
        NormalizeColumns.protein(df)
        _get_shortest_name(
            df, "synonyms"
        )  # Take the shortest name in protein names list as name
        if not df.index.is_numeric():
            df = df.reset_index().copy()
        df = df[~df[self._id_field].isnull()]
        return df.set_index(self._id_field)



"""Developer API.

.. autosummary::
   :toctree: .

    ontology_info
   check_if_index_compliant
   get_compliant_index_from_column
"""
from ._fix_index import (
    check_if_index_compliant,
    get_compliant_index_from_column,
    explode_aggregated_column_to_expand,
)  # noqa

from ._ontology import ontology_info
from ._current_db_version import current_db_version



import shutil
from pathlib import Path

from .dev._io import load_yaml, write_yaml

ROOT = Path(__file__).parent / "versions"
VERSIONS = ROOT / "versions.yaml"
_LOCAL = ROOT / "_local.yaml"
_CURRENT = ROOT / "_current.yaml"
_DB = ROOT / "_lndb.yaml"

versions = load_yaml(VERSIONS)


# if _local.yaml doesn't exist, copy from versions.yaml
if not _LOCAL.exists():

    def write_local_yaml(versions):
        """Make sure version keys are strings."""
        _local = {}
        for name, db_versions in versions.items():
            db = next(iter(db_versions))
            versions = db_versions.get(db).get("versions")
            _local[name] = {db: {"versions": {}}}
            for v, v_url in versions.items():
                _local[name][db]["versions"][str(v)] = v_url
        return _local

    _local = write_local_yaml(versions)
    write_yaml(_local, _LOCAL)


_local = load_yaml(_LOCAL)

# update _local to add additional entries from the public versions.yaml table
for entity, dbs in versions.items():
    if entity not in _local:
        _local[entity] = versions[entity]
    else:
        for db_name, v in dbs.items():
            if db_name not in _local[entity]:
                _local[entity][db_name] = dbs[db_name]
            else:
                for version in v["versions"]:
                    if version not in _local[entity][db_name]["versions"]:
                        _local[entity][db_name]["versions"][version] = v["versions"][
                            version
                        ]

# writes the most recent version to the _current.yaml
if not _CURRENT.exists():

    def write_current_yaml(versions):
        _current = {}
        for name, db_versions in versions.items():
            # this will only take the 1st db if multiple exists for the same entity
            db = next(iter(db_versions))
            versions = db_versions.get(db).get("versions")
            version = str(sorted(versions.keys(), reverse=True)[0])
            _current[name] = {db: version}
        return _current

    _current = write_current_yaml(versions)
    write_yaml(_current, _CURRENT)

# if no _lndb file, write _current to _lndb for lndb_setup
if not _DB.exists():
    shutil.copy2(_CURRENT, _DB)



from typing import Optional

import pandas as pd
from cached_property import cached_property

from .._normalize import GENE_COLUMNS, NormalizeColumns
from .._settings import s3_bionty_assets
from .._table import EntityTable

ALIAS_DICT = {"symbol": "synonyms"}
FILENAMES = {
    "human": "KJ1HgB695AqbVWvfit8sl.parquet",
    "mouse": "xaBDkhBYLXWHq6gJYnedD.parquet",
}


class Gene(EntityTable):
    """Gene.

    The default indexer is `ensembl_gene_id`

    Args:
        species: `name` of `Species` entity EntityTable.
        id: default is `ensembl_gene_id`

    Notes:
        Biotypes: https://www.ensembl.org/info/genome/genebuild/biotypes.html
        Gene Naming: https://www.ensembl.org/info/genome/genebuild/gene_names.html

    """

    def __init__(
        self,
        species: str = "human",
        id: Optional[str] = None,
        database: Optional[str] = None,
        version: Optional[str] = None,
    ):
        super().__init__(id=id, database=database, version=version)
        if FILENAMES.get(species) is None:
            raise NotImplementedError
        self._species = species
        self._id_field = "ensembl_gene_id" if id is None else id
        self._lookup_col = "symbol"

    @property
    def species(self):
        """The `name` of `Species` entity EntityTable."""
        return self._species

    @cached_property
    def df(self):
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/ensembl-gene
        """
        cloudpath = s3_bionty_assets(FILENAMES.get(self.species))
        self._filepath = cloudpath.fspath

        df = pd.read_parquet(self._filepath)
        NormalizeColumns.gene(df, species=self.species)
        if not df.index.is_numeric():
            df = df.reset_index().copy()
        df = df[~df[self._id_field].isnull()]
        return df.set_index(self._id_field)

    def curate(  # type: ignore
        self, df: pd.DataFrame, column: str = None
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier.

        - If `column` is `None`, checks the existing index for compliance with
          the default identifier.
        - If `column` denotes an entity identifier, tries to map that identifier
          to the default identifier.

        Returns the DataFrame with the curated index and a boolean `__curated__`
        column that indicates compliance with the default identifier.

        In addition to the .curate() in base class, this also performs alias mapping.
        """
        agg_col = ALIAS_DICT.get(self._id_field)
        df = df.copy()

        # if the query column name does not match any columns in the self.df
        # Bionty assume the query column and the self._id_field uses the same type of
        # identifier
        orig_column = column
        if column is not None and column not in self.df.columns:
            # normalize the identifier column
            column_norm = GENE_COLUMNS.get(column)
            if column_norm in df.columns:
                raise ValueError("{column_norm} column already exist!")
            else:
                column = self._id_field if column_norm is None else column_norm
                df.rename(columns={orig_column: column}, inplace=True)
            agg_col = ALIAS_DICT.get(column)

        return (
            super()
            ._curate(df=df, column=column, agg_col=agg_col)
            .rename(columns={column: orig_column})
        )



from typing import Optional

import pandas as pd
from cached_property import cached_property

from .._table import EntityTable

SPECIES_FILENAME = "VpdUdouFahpvStwddqTwk.parquet"


class Species(EntityTable):
    """Species.

    Args:
        id: Field name that should constitute the primary reference for each
            value. It will also be the primary key in the corresponding SQL EntityTable.
    """

    def __init__(
        self,
        id: Optional[str] = None,
        database: Optional[str] = None,
        version: Optional[str] = None,
    ):
        super().__init__(id=id, database=database, version=version)
        self._id_field = "name" if id is None else id
        self._lookup_col = "name"

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/ensembl-species
        """
        url = f"https://ftp.ensembl.org/pub/{self._version}/species_EnsemblVertebrates.txt"  # noqa
        self._filepath = self._url_download(url)

        df = pd.read_csv(self._filepath, sep="\t", index_col=False)
        df.rename(
            columns={
                "#name": "name",
                "species": "scientific_name",
                "taxonomy_id": "taxon_id",
            },
            inplace=True,
        )
        df["name"] = df["name"].str.lower()
        df.insert(0, "id", "NCBI_" + df["taxon_id"].astype(str))

        return df.set_index(self._id_field)



import pytest

from bionty import CellMarker, Gene, Protein
from bionty._table import EntityTable, _camel_to_snake


def test_entity_table():
    entity_table = EntityTable()
    with pytest.raises(NotImplementedError):
        entity_table.df
    assert entity_table.entity == "entity_table"
    assert _camel_to_snake("EntityTable") == "entity_table"


def species_not_implemented():
    with pytest.raises(NotImplementedError):
        CellMarker(species="mouse")
    with pytest.raises(NotImplementedError):
        Protein(species="cat")
    with pytest.raises(NotImplementedError):
        Gene(species="cat")



import re
from collections import namedtuple
from pathlib import Path
from typing import Iterable, NamedTuple, Optional

import pandas as pd
from cached_property import cached_property

from ._logger import logger
from ._ontology import Ontology
from ._settings import check_dynamicdir_exists, settings
from .dev._fix_index import (
    check_if_index_compliant,
    explode_aggregated_column_to_expand,
    get_compliant_index_from_column,
)
from .dev._io import load_yaml, url_download

VERSIONS_PATH = Path(__file__).parent / "versions"


def _camel_to_snake(string: str) -> str:
    """Convert CamelCase to snake_case."""
    return re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower()


class EntityTable:
    """Biological entity as a EntityTable.

    See :doc:`guide/index` for background.
    """

    def __init__(
        self,
        id: Optional[str] = None,
        database: Optional[str] = None,
        version: Optional[str] = None,
    ):
        self._id_field = "id" if id is None else id
        # By default lookup allows auto-completion for name and returns the id.
        # lookup column can be changed using `.lookup_col = `.
        self._lookup_col = "name"
        if self.__class__.__name__ == "EntityTable":
            return None
        self._get_version(database=database, version=version)

    @cached_property
    def entity(self) -> str:
        """Name of the entity."""
        return _camel_to_snake(self.__class__.__name__)

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame representation of EntityTable."""
        raise NotImplementedError

    @property
    def lookup_col(self) -> str:
        """The column that allows auto-completion."""
        return self._lookup_col

    @lookup_col.setter
    def lookup_col(self, column_name) -> None:
        self._lookup_col = column_name

    @cached_property
    def lookup(self) -> NamedTuple:
        """Return an auto-complete object for the bionty id."""
        df = self.df.reset_index()
        if self._lookup_col not in df:
            raise AssertionError(f"No {self._lookup_col} column exists!")

        # uniquefy lookup keys
        df.index = self._uniquefy_duplicates(
            self._to_lookup_keys(df[self._lookup_col].values)
        )

        return self._namedtuple_from_dict(df[self._id_field].to_dict())

    def _to_lookup_keys(self, x: list) -> list:
        """Convert a list of strings to tab-completion allowed formats."""
        lookup = [re.sub("[^0-9a-zA-Z]+", "_", str(i)) for i in x]
        for i, value in enumerate(lookup):
            if not value[0].isalpha():
                lookup[i] = f"LOOKUP_{value}"
        return lookup

    def _namedtuple_from_dict(
        self, mydict: dict, name: Optional[str] = None
    ) -> NamedTuple:
        """Create a namedtuple from a dict to allow autocompletion."""
        if name is None:
            name = self.entity
        nt = namedtuple(name, mydict)  # type:ignore
        return nt(**mydict)

    def _uniquefy_duplicates(self, lst: Iterable) -> list:
        """Uniquefy duplicated values in a list."""
        df = pd.DataFrame(lst)
        duplicated = df[df[0].duplicated(keep=False)]
        df.loc[duplicated.index, 0] = (
            duplicated[0] + "__" + duplicated.groupby(0).cumcount().astype(str)
        )
        return list(df[0].values)

    def _ontology_to_df(self, ontology: Ontology):
        """Convert ontology to a DataFrame with id and name columns."""
        return pd.DataFrame(
            [(term.id, term.name) for term in ontology.terms()],
            columns=["ontology_id", "name"],
        ).set_index(self._id_field)

    def _curate(
        self, df: pd.DataFrame, column: str = None, agg_col: str = None
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier."""
        df = df.copy()

        if agg_col is not None:
            # if provided a column with aggregated values, performs alias mapping
            alias_map = explode_aggregated_column_to_expand(
                self.df.reset_index(), aggregated_col=agg_col, target_col=self._id_field
            )[self._id_field]

        if column is None:
            # when column is None, use index as the input column
            index_name = df.index.name
            df["__mapped_index"] = (
                df.index if agg_col is None else df.index.map(alias_map)
            )
            df["orig_index"] = df.index
            df.index = df["__mapped_index"].fillna(df["orig_index"])
            del df["__mapped_index"]
            df.index.name = index_name
            matches = check_if_index_compliant(df.index, self.df.index)
        else:
            orig_series = df[column]
            df[column] = df[column] if agg_col is None else df[column].map(alias_map)
            df[column] = df[column].fillna(orig_series)
            new_index, matches = get_compliant_index_from_column(
                df=df,
                ref_df=self.df,
                column=column,
            )

            # keep the original index name as column name if exists
            # otherwise name it "orig_index"
            if df.index.name is None:
                df["orig_index"] = df.index
            else:
                df[df.index.name] = df.index
            df.index = new_index
            df.index.name = self._id_field
            df[column] = orig_series.values  # keep the original column untouched
        # annotated what complies with the default ID
        df["__curated__"] = matches
        # some stats for logging
        n_misses = len(matches) - matches.sum()
        frac_misses = round(n_misses / len(matches) * 100, 1)
        n_mapped = matches.sum()
        frac_mapped = 100 - frac_misses
        logger.success(f"{n_mapped} terms ({frac_mapped}%) are linked.")
        logger.warning(f"{n_misses} terms ({frac_misses}%) are not linked.")
        return df

    @check_dynamicdir_exists
    def _url_download(self, url: str):
        """Download file from url to dynamicdir."""
        filename = url.split("/")[-1]
        if not self._localpath(filename).exists():
            logger.info(
                f"Downloading {self.entity} reference for the first time might take a"
                " while..."
            )
            url_download(url, self._localpath(filename))
        return self._localpath(filename)

    def _ontology_localpath_from_url(self, url: str):
        """Get version from the ontology url."""
        version = url.split("/")[-2]
        filename = url.split("/")[-1]
        return settings.dynamicdir / f"{version}___{filename}"

    def _load_current_version(self):
        """Load current version."""
        try:
            import lndb_setup

            db = lndb_setup.settings._instance_exists
        except ImportError:
            db = False

        if db:
            filename = "_lndb.yaml"
        else:
            filename = "_current.yaml"

        ((database, version),) = (
            load_yaml(VERSIONS_PATH / filename).get(self.__class__.__name__).items()
        )

        return database, version

    def _load_versions(self):
        """Load all versions with string version keys."""
        versions = load_yaml(VERSIONS_PATH / "versions.yaml").get(
            self.__class__.__name__
        )

        versions_db = {}

        for db, vers in versions.items():
            versions_db[db] = {"versions": {}}
            for k in vers["versions"]:
                versions_db[db]["versions"][str(k)] = versions[db]["versions"][k]
        return versions_db

    def _get_version(
        self, database: Optional[str] = None, version: Optional[str] = None
    ):
        # Read in all the versions from the versions.yaml file.
        database_, version_ = self._load_current_version()
        db_versions = self._load_versions()
        # Use the latest version if version is None.
        self._database = database_ if database is None else database
        self._version = version_ if version is None else str(version)
        self._url = db_versions.get(self._database).get("versions").get(self._version)
        if self._url is None:
            raise ValueError(
                f"Databse {self._database} version {self._version} is not found, select"
                f" one of the following: {db_versions}"
            )

    def _localpath(self, filename: str):
        """Return the local path of a filename marked with version."""
        return settings.dynamicdir / f"{self._version}___{filename}"

    def curate(self, df: pd.DataFrame, column: str = None, case_sensitive: bool = True):
        """Curate index of passed DataFrame to conform with default identifier.

        - If `column` is `None`, checks the existing index for compliance with
          the default identifier.
        - If `column` denotes an entity identifier, tries to map that identifier
          to the default identifier.

        Returns the DataFrame with the curated index and a boolean `__curated__`
        column that indicates compliance with the default identifier.
        """
        df = df.copy()
        orig_column = column
        if column is not None and column not in self.df.columns:
            column = self._id_field
            df.rename(columns={orig_column: column}, inplace=True)

        if not case_sensitive:
            if column in df.columns:
                df[column] = df[column].str.upper()
            else:
                df.index = df.index.str.upper()
        return self._curate(df=df, column=column).rename(columns={column: orig_column})

    def ontology(self, **kwargs) -> Ontology:
        """Ontology."""
        # Get the in-use url from yaml file
        localpath = self._url_download(self._url)

        return Ontology(handle=localpath, **kwargs)



from typing import Optional

import pandas as pd
from cached_property import cached_property

from .._ontology import Ontology
from .._settings import settings
from .._table import EntityTable


class CellType(EntityTable):
    """Cell type.

    Edits of terms are coordinated and reviewed on:
    https://github.com/obophenotype/cell-ontology
    """

    def __init__(
        self,
        id: str = "ontology_id",
        database: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None:
        database = "cl" if database is None else database
        super().__init__(id=id, database=database, version=version)

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame."""
        self._filepath = settings.datasetdir / "celltype_lookup.parquet"

        if not self._filepath.exists():
            df = self._ontology_to_df(self.ontology)
            df.to_parquet(self._filepath)

        return pd.read_parquet(self._filepath).reset_index().set_index(self._id_field)

    @cached_property
    def ontology(self) -> Ontology:  # type:ignore
        """Cell ontology."""
        return super().ontology()



from bionty import lookup


def test_lookup():
    assert lookup.gene_id.ensembl_gene_id == "ensembl_gene_id"
    assert lookup.protein_id.uniprotkb_id == "uniprotkb_id"
    assert lookup.species.mouse == "mouse"



import pandas as pd

GENE_COLUMNS = {
    "Gene stable ID": "ensembl_gene_id",
    "Transcript stable ID": "ensembl_transcript_id",
    "Protein stable ID": "ensembl_protein_id",
    "Gene name": "symbol",
    "Gene type": "gene_type",
    "Gene description": "description",
    "NCBI gene (formerly Entrezgene) ID": "ncbi_gene_id",
    "HGNC ID": "hgnc_id",
    "MIM gene accession": "omim_id",
    "Gene Synonym": "synonyms",
    "MGI ID": "mgi_id",
}

PROTEIN_COLUMNS = {
    "Entry": "uniprotkb_id",
    "Entry Name": "uniprotkb_name",
    "Protein names": "synonyms",
    "Length": "length",
    "Organism (ID)": "species_id",
    "Gene Names (primary)": "gene_symbols",
    "Gene Names (synonym)": "gene_synonyms",
    "Ensembl": "ensembl_transcript_ids",
    "GeneID": "ncbi_gene_ids",
}


class NormalizeColumns:
    """Standardizing column names."""

    @staticmethod
    def gene(df: pd.DataFrame, species=None):
        """Column names of gene EntityTables."""
        df.rename(columns=GENE_COLUMNS, inplace=True)

    @staticmethod
    def protein(df: pd.DataFrame, species=None):
        """Column names of protein EntityTables."""
        df.rename(columns=PROTEIN_COLUMNS, inplace=True)



"""Bionty: Data model generator for biology.

Import the package::

   import bionty as bt

Entities:

.. autosummary::
   :toctree: .

   Gene
   Protein
   Species
   CellType
   CellMarker
   Tissue
   Disease
   Readout

The base model for every entity class is:

.. autosummary::
   :toctree: .

   EntityTable

Lookup of vocabulary:

.. autosummary::
   :toctree: .

   lookup

External API:

.. autosummary::
   :toctree: .

   Ontology

"""

__version__ = "0.6.5"

# prints warning of python versions
from lamin_logger import py_version_warning

py_version_warning("3.7", "3.10")

from . import _sync_versions

# dynamic classes
from ._gene import Gene
from ._species import Species
from ._protein import Protein
from ._celltype import CellType
from ._cellmarker import CellMarker
from ._tissue import Tissue
from ._disease import Disease
from ._readout import Readout

# tools
from ._normalize import NormalizeColumns
from ._table import EntityTable
from ._ontology import Ontology
from ._lookup import lookup



import pandas as pd


def current_db_version(db: str):
    """Lookup the current version of a database."""
    db = db.lower()

    if db == "ensembl":
        # For Ensembl, parse the current_README file
        lines = []

        for line in pd.read_csv(
            "https://ftp.ensembl.org/pub/current_README",
            chunksize=1,
            header=None,
            encoding="utf-8",
        ):
            lines.append(line.iloc[0, 0])
        return "-".join(lines[1].split(" ")[1:-1]).lower()

    else:
        raise NotImplementedError
